#ifndef OBJ_H
#define OBJ_H

#include "aliases.h"
#include "img.h"
#include "mem_arena.h"
#include "typed_list.h"
#include "vec.h"

#define INVALID_MODEL ((Model){0})
#define IS_INVALID_MODEL(m) (m.vertices == NULL || m.triangles == NULL)

typedef struct triangle Triangle;
struct triangle {
  u64 p0;
  u64 p1;
  u64 p2;
  u64 n0;
  u64 n1;
  u64 n2;
  u64 tx0;
  u64 tx1;
  u64 tx2;
};

typedef enum {
  RENDER_TYPE_WIREFRAME,
  RENDER_TYPE_FILLED,
  RENDER_TYPE_SHADED,

  COUNT_RENDER_TYPES,
} RenderType;

typedef enum {
  PROJECTION_TYPE_ORTHOGRAPHIC,
  PROJECTION_TYPE_PERSPECTIVE,

  COUNT_PROJECTION_TYPE,
} ProjectionType;

MAKE_LIST_TYPE(Triangle);

typedef struct model Model;
struct model {
  LIST_TYPE(V3f) * vertices;
  LIST_TYPE(V3f) * normals;
  LIST_TYPE(V2f) * texture_coordinates;
  LIST_TYPE(Triangle) * triangles;
  Image *texture;
};

typedef struct render Render;
struct render {
  Image img;
  Depth depth;
};

typedef V3f(VertexShader)(const V3f *vertex, M4x4f *model_view,
                          M4x4f *projection, const Render *render);
typedef bool(FragmentShader)(Colour colour);

typedef struct shader Shader;
struct shader {
  VertexShader *vertex;
  FragmentShader *fragment;
  M4x4f *model_view;
  M4x4f *projection;
};

Model load_obj_file(Arena *arena, const char *filename, const char *texture);
bool init_render(Arena *arena, Render *render, u64 width, u64 height);
void render_model(const Model *model, Render *render, const Shader *shader,
                  RenderType render_type, Colour colour);

#endif // OBJ_H
